<?php

/*
 * FATEK communication protocol implementation
 *
 * Last modified: May 29, 2019
 *
 * Copyright (c) 2019 Mariusz Zynel & Grzegorz Lewandowski.
 *
 * This  software  is  FREE.  You can use  and/or  redistribute  it for any
 * purpose in either,  modified, or unmodified form, under the terms of the
 * GNU General Public License as published by the Free Software Foundation.
 *
 * The above copyright notice and this permission  notice shall be included
 * in all copies or substantial portions of this software.
 *
 * THIS  SOFTWARE IS PROVIDED AS IS AND COME WITH NO WARRANTY  OF ANY KIND,
 * EITHER  EXPRESSED OR IMPLIED.  IN NO EVENT WILL THE COPYRIGHT  HOLDER BE
 * LIABLE FOR ANY DAMAGES RESULTING FROM THE USE OF THIS SOFTWARE
 */

require_once('stdlib.php');

define('FATEK_OK',                0);

define('FATEK_BUF_SIZE',          1024);

define('FATEK_MSG_START_CODE',    0x02);
define('FATEK_MSG_END_CODE',      0x03);

define('FATEK_CMD_GET_STATUS',    0x40);
define('FATEK_CMD_SET_RUNSTOP',   0x41);
define('FATEK_CMD_SET_STATE',     0x42);
define('FATEK_CMD_GET_STATE',     0x43);
define('FATEK_CMD_GET_BVALUE',    0x44);
define('FATEK_CMD_SET_BVALUE',    0x45);
define('FATEK_CMD_GET_IVALUE',    0x46);
define('FATEK_CMD_SET_IVALUE',    0x47);
define('FATEK_CMD_GET_VALUE',     0x48);
define('FATEK_CMD_SET_VALUE',     0x49);
define('FATEK_CMD_LOOPBACK',      0x4E);
define('FATEK_CMD_READ_PROGRAM',  0x4F);
define('FATEK_CMD_WRITE_PROGRAM', 0x50);
define('FATEK_CMD_GET_DETAILS',   0x53);

define('FATEK_REG_TYPE_BOOL',     0x01);
define('FATEK_REG_TYPE_16BIT',    0x10);
define('FATEK_REG_TYPE_32BIT',    0x20);

define('FATEK_REG_SYMBOL_X',      0x58);
define('FATEK_REG_SYMBOL_Y',      0x59);
define('FATEK_REG_SYMBOL_M',      0x4D);
define('FATEK_REG_SYMBOL_S',      0x53);
define('FATEK_REG_SYMBOL_T',      0x54);
define('FATEK_REG_SYMBOL_C',      0x43);
define('FATEK_REG_SYMBOL_R',      0x52);
define('FATEK_REG_SYMBOL_D',      0x44);

define('FATEK_REG_STATE_DISABLE', 0x01);
define('FATEK_REG_STATE_ENABLE',  0x02);
define('FATEK_REG_STATE_SET',     0x03);
define('FATEK_REG_STATE_RESET',   0x04);

function hex_dump($data, $newline = "\n") {
    static $from = '';
    static $to = '';
    static $width = 16; # number of bytes per line
    static $pad = '.';  # padding for non-visible characters

    if ($from === '') {
        for ($i=0; $i <= 0xFF; $i++) {
            $from .= chr($i);
            $to .= ($i >= 0x20 && $i <= 0x7E) ? chr($i) : $pad;
        }
    }

    $hex = str_split(bin2hex($data), $width * 2);
    $chars = str_split(strtr($data, $from, $to), $width);

    $offset = 0;
    foreach ($hex as $i => $line) {
        echo sprintf('%02X : %-47s [%s]', $offset, implode(' ', str_split($line, 2)), $chars[$i]) . $newline;
        $offset += $width;
    }
}

class FatekMessage {
    public $station;
    public $command;
    public $error;
    public $data;
    public $size;
}

class FatekRegister {
    public $type;
    public $symbol;
    public $address;
    public $state;
    public $value;
    private $format_bool = array(
        'X'   => 'X%04d',
        'Y'   => 'Y%04d',
        'M'   => 'M%04d',
        'S'   => 'S%04d',
        'T'   => 'T%04d',
        'C'   => 'C%04d'
    );
    private $format_16bit = array(
        'X'   => 'WX%04d',
        'Y'   => 'WY%04d',
        'M'   => 'WM%04d',
        'S'   => 'WS%04d',
        'T'   => 'RT%04d',
        'C'   => 'RC%04d',
        'R'   => 'R%05d',
        'D'   => 'D%05d'
    );
    private $format_32bit = array(
        'X'   => 'DWX%04d',
        'Y'   => 'DWY%04d',
        'M'   => 'DWM%04d',
        'S'   => 'DWS%04d',
        'T'   => 'DRT%04d',
        'C'   => 'DRC%04d',
        'R'   => 'DR%05d',
        'D'   => 'DD%05d'
    );

    public function __construct($type, $symbol, $address, $value = 0) {
        $this->type = $type;
        $this->symbol = $symbol;
        $this->address = $address;
        $this->value = $value;
    }

    public function details() {
        return sprintf('type: %d symbol: %c address: %d value: %s', $this->type, $this->symbol, $this->address, $this->value);
    }

    public function  format_symbol_address() {
        switch ($this->type) {
            case FATEK_REG_TYPE_BOOL:
                return sprintf($this->format_bool[chr($this->symbol)], $this->address);
            case FATEK_REG_TYPE_16BIT:
                return sprintf($this->format_16bit[chr($this->symbol)], $this->address);
            case FATEK_REG_TYPE_32BIT:
                return sprintf($this->format_32bit[chr($this->symbol)], $this->address);
            default:
                throw new ExceptionRoot('Nieznany typ rejestru.');
        }
    }

    public function format_symbol_address_value() {
        switch ($this->type) {
            case FATEK_REG_TYPE_BOOL:
                return sprintf("%s%1X", $this->format_symbol_address(), $this->value);
            case FATEK_REG_TYPE_16BIT:
                return sprintf("%s%04X", $this->format_symbol_address(), $this->value);
            case FATEK_REG_TYPE_32BIT:
                return sprintf("%s%08X", $this->format_symbol_address(), $this->value);
            default:
                throw new ExceptionRoot('Nieznany typ rejestru.');
        }
    }
}

class FatekPLC {
    protected $socket;
    protected $host;
    protected $port;
    private $debug;

    public function __construct($host = PLC_HOST, $port = PLC_PORT, $debug = PLC_DEBUG) {
        $this->socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
        $this->host = $host;
        $this->port = $port;
        $this->debug = $debug;
    }

    public function connect($timeout = 10) {
        socket_connect($this->socket, $this->host, $this->port);
    }

    public function disconnect() {
        socket_close($this->socket);
    }

    protected function read() {
        $buf = $chunk = NULL;
        while ((is_null($chunk) || bin2hex($chunk[strlen($chunk) - 1]) == chr(FATEK_MSG_END_CODE))
                && $chunk = socket_read($this->socket, FATEK_BUF_SIZE)) {
            $buf .= $chunk;
        }
        return $buf;
    }

    protected function write($buf) {
        socket_write($this->socket, $buf);
    }

    protected function lrc($buf) {
        $lrc = 0;

        for ($i = 0; $i < strlen($buf); $i++) {
            $lrc = ($lrc + hexdec(bin2hex($buf[$i]))) & 0xFF;
        }

        return $lrc;
    }

    protected function message(FatekMessage $req, FatekMessage &$rsp) {
        $offset = 0;

        $buf = sprintf('%c%02X%02X%s', FATEK_MSG_START_CODE, $req->station, $req->command, $req->data);
        $buf .= sprintf('%02X%c', $this->lrc($buf), FATEK_MSG_END_CODE);

        if ($this->debug) {
            echo sprintf("message: station: %d command: %d\n", $req->station, $req->command);
            echo sprintf("message: write: %d bytes\n", strlen($buf));
            hex_dump($buf);
        }

        $this->write($buf);

        $buf = $this->read();

        if ($this->debug) {
            echo sprintf("message: read: %d bytes\n", strlen($buf));
            hex_dump($buf);
        }

        if ($req->command == FATEK_CMD_LOOPBACK) {
            if (sscanf($buf, '%c%02X%02X', $s, $rsp->station, $rsp->command) != 3)
                throw new ExceptionRoot('Nieprawidowa odpowied.');
        } else {
            if (sscanf($buf, '%c%02X%02X%1X', $s, $rsp->station, $rsp->command, $rsp->error) != 4)
                throw new ExceptionRoot('Nieprawidowa odpowied.');
        }

        $rsp->size = strlen($buf) - 9;
        $rsp->data = substr($buf, 6, $rsp->size);

        if (hexdec(substr($buf, -3, 2)) != $this->lrc(substr($buf, 0, -3)))
            throw new ExceptionRoot('Niezgodna suma kontrolna.');

        return FATEK_OK;
    }

    public function get_status($station) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("get_status: station: %d\n", $station);

        $req->station = $station;
        $req->command = FATEK_CMD_GET_STATUS;
        $req->data = NULL;
        $req->size = 0;

        $this->message($req, $rsp);

        foreach (str_split($rsp->data, 2) as $i => $hex)
            printf("data[%03d]: %3d %s %s\n", $i, hexdec($hex), $hex,
                implode(' ', str_split(str_pad(base_convert($hex, 16, 2), 8, "0", STR_PAD_LEFT), 4)));

        return FATEK_OK;
    }

    public function runstop($station, $ccode) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("runstop: station: %d ccode: %d\n", $station, $ccode);

        $req->station = $station;
        $req->command = FATEK_CMD_SET_RUNSTOP;
        $req->data = sprintf('%1X', $ccode);
        $req->size = 1;

        return $this->message($req, $rsp);
    }

    public function set_state($station, FatekRegister $reg) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("set_state: station: %d symbol: %c address: %d state: %d\n", $station, $reg->symbol, $reg->address, $reg->state);

        $req->station = $station;
        $req->command = FATEK_CMD_SET_STATE;
        $req->data = sprintf('%1X%s', $reg->state, $reg->format_symbol_address());
        $req->size = strlen($req->data);

        return $this->message($req, $rsp);
    }

    public function get_state($station, array &$regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("get_state: station: %d symbol: %c address: %d number: %d\n", $station, $regs[0]->symbol, $regs[0]->address, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_GET_STATE;
        $req->data = sprintf('%02X%s', count($regs), $regs[0]->format_symbol_address());
        $req->size = strlen($req->data);

        $this->message($req, $rsp);

        if ($rsp->size != count($regs))
            throw new ExceptionRoot('Nieprawidowa ilo rejestrw w odpowiedzi.');

        foreach ($regs as $i => $reg) {
            $reg->type = FATEK_REG_TYPE_BOOL;
            $reg->symbol = $regs[0]->symbol;
            $reg->address = $regs[0]->address + $i;
            $reg->state = hexdec($rsp->data[$i]);
            if ($this->debug)
                echo sprintf("%s state: %d\n", $reg->format_symbol_address(), $reg->state);
        }

        return FATEK_OK;
    }

    public function get_bvalue($station, array $regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("get_bvalue: station: %d symbol: %c address: %d number: %d\n", $station, $regs[0]->symbol, $regs[0]->address, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_GET_BVALUE;
        $req->data = sprintf('%02X%s', count($regs), $regs[0]->format_symbol_address());
        $req->size = strlen($req->data);

        $this->message($req, $rsp);

        if ($rsp->size != count($regs))
            throw new ExceptionRoot('Nieprawidowy rozmiar odpowiedzi.');

        foreach ($regs as $i => $reg) {
            $reg->type = FATEK_REG_TYPE_BOOL;
            $reg->symbol = $regs[0]->symbol;
            $reg->address = $regs[0]->address + $i;
            $reg->value = hexdec($rsp->data[$i]);
            if ($this->debug)
                echo sprintf("%c%02d = %d\n", $reg->symbol, $reg->address, $reg->value);
        }

        return FATEK_OK;
    }

    public function set_bvalue($station, array &$regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("set_bvalue: station: %d symbol: %c address: %d number: %d\n", $station, $regs[0]->symbol, $regs[0]->address, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_SET_BVALUE;
        $req->data = sprintf('%02X%s', count($regs), $regs[0]->format_symbol_address());
        foreach ($regs as $reg)
            $req->data .= sprintf('%1X', $reg->value);
        $req->size = strlen($req->data);

        return $this->message($req, $rsp);
    }

    public function get_ivalue($station, array $regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("get_ivalue: station: %d symbol: %c address: %d number: %d\n", $station, $regs[0]->symbol, $regs[0]->address, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_GET_IVALUE;
        $req->data = sprintf('%02X%s', count($regs), $regs[0]->format_symbol_address());
        $req->size = strlen($req->data);

        $this->message($req, $rsp);

        if ($regs[0]->type == FATEK_REG_TYPE_32BIT)
            $chars = 8;
        else
            $chars = 4;

        if ($rsp->size != count($regs) * $chars)
            throw new ExceptionRoot('Nieprawidowy rozmiar odpowiedzi.');

        foreach ($regs as $i => $reg) {
            $reg->type = $regs[0]->type;
            $reg->symbol = $regs[0]->symbol;
            $reg->address = $regs[0]->address + $i * $chars / 4;
            $reg->value = hexdec(substr($rsp->data, $i * $chars, $chars));
            if ($this->debug)
                echo sprintf("%s = %lu\n", $reg->format_symbol_address(), $reg->value);
        }

        return FATEK_OK;
    }

    public function set_ivalue($station, array $regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("set_ivalue: station: %d symbol: %c address: %d number: %d\n", $station, $regs[0]->symbol, $regs[0]->address, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_SET_IVALUE;
        $req->data = sprintf('%02X%s', count($regs), $regs[0]->format_symbol_address());
        $req->size = strlen($req->data);

        if ($regs[0]->type == FATEK_REG_TYPE_32BIT)
            $format = '%08X';
        else
            $format = '%04X';
        foreach ($regs as $i => $reg)
            $req->data .= sprintf($format, $reg->value);

        return $this->message($req, $rsp);
    }

    public function get_value($station, array $regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("get_value: station: %d number: %d\n", $station, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_GET_VALUE;
        $req->data = sprintf('%02X', count($regs));

        foreach ($regs as $reg)
            $req->data .= $reg->format_symbol_address();
        $req->size = strlen($req->data);

        $this->message($req, $rsp);

        $offset = 0;
        foreach ($regs as $reg) {
            switch ($reg->type) {
                case FATEK_REG_TYPE_BOOL:
                    if (sscanf(substr($rsp->data, $offset, 1), '%1X', $reg->value) != 1)
                        throw new ExceptionRoot('Nieprawidowa warto bool w odpowiedzi.', 0, $reg->details());
                    $offset += 1;
                    break;
                case FATEK_REG_TYPE_16BIT:
                    if (sscanf(substr($rsp->data, $offset, 4), '%04X', $reg->value) != 1)
                        throw new ExceptionRoot('Nieprawidowa warto 16bit w odpowiedzi.', 0, $reg->details());
                    $offset += 4;
                    break;
                case FATEK_REG_TYPE_32BIT:
                    if (sscanf(substr($rsp->data, $offset, 8), '%08X', $reg->value) != 1)
                        throw new ExceptionRoot('Nieprawidowa warto 32bit w odpowiedzi.', 0, $reg->details());
                    $offset += 8;
                    break;
                default:
                    throw new ExceptionRoot('Nieznany typ rejestru.');
            }
            if ($this->debug)
                echo sprintf("%s = %lu\n", $reg->format_symbol_address(), $reg->value);
        }

        return FATEK_OK;
    }

    public function set_value($station, array $regs) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        if ($this->debug)
            echo sprintf("set_value: station: %d number: %d\n", $station, count($regs));

        $req->station = $station;
        $req->command = FATEK_CMD_SET_VALUE;
        $req->data = sprintf('%02X', count($regs));

        foreach ($regs as $reg)
            $req->data .= $reg->format_symbol_address_value();
        $req->size = strlen($req->data);

        return $this->message($req, $rsp);
    }

    public function loopback($station, $str = NULL) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        $req->station = $station;
        $req->command = FATEK_CMD_LOOPBACK;
        $req->data = is_null($str) ? 'TEST abcdefghijklmnopqrstuvwxz 0123456789' : $str;
        $req->size = strlen($this->data);

        $this->message($req, $rsp);

        return FATEK_OK;
    }
    
    public function get_details($station) {
        $req = new FatekMessage();
        $rsp = new FatekMessage();

        $req->station = $station;
        $req->command = FATEK_CMD_GET_DETAILS;
        $req->data = NULL;
        $req->size = 0;

        $this->message($req, $rsp);

        foreach (str_split($rsp->data, 2) as $i => $hex)
            printf("data[%03d]: %3d %s %s\n", $i, hexdec($hex), $hex, 
                implode(' ', str_split(str_pad(base_convert($hex, 16, 2), 8, "0", STR_PAD_LEFT), 4)));

        return FATEK_OK;
    }
}

?>
